 /* BBSread version 1.3  by C.A.B. */
 /* Original version Jan 1994, updated Jan 1995 to fix bugs and introduce */
 /* a binary tree and linked list. */
 /* Updated again November 1995 to fix errors and add Linux and Windows */
 /* to the list */
 /* Reads the BBS listings to create statistics on software use etc */
 /* Run it with 'test' as a command line parameter to display unknown */
 /* computer types and other test stuff */

 /* Baud rates and the speed in bps at which data is transmitted using them: */
 /* (To the best of my knowledge - might not be 100% accurate) */
 /* v.34bis = 34000 v.34 & v.fc = 28800  turbopep = 23000  v.32terbo = 19200, v.32 = 14400, */
 /* v.32bis & pep  & v.42bis  v32 & hst = 9600 v.22bis = 2400 */
 /* v.23 = 1200/75 v.22 = 1200  v.21 = 300 */

#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <graphics.h>
#include <stdlib.h>
#include <malloc.h>
#include <io.h>
#include <fcntl.h>
#include <time.h>

#define maxbaud 11
#define yes 1
#define no 0
#define maxsw 50
#define maxcomp 9
#define maxibm 6
#define alphabetical 1         /* To determine sorting order */

void foundcomp(char [], int);
void foundsw(char []);
void foundbaud(char []);
void foundhours(char []);
int checkfors();
void intro();
void writeresults();
int find(char[],char[]);
long getsizes(char []);
void drawit();
void show (float);
void last();
void showfile();
void writepercent(FILE *,int,int,char[]);
void initialise ();
void testing ();
struct llist *addtotree (struct llist *, char *);
struct llist *addtolist (struct llist *, char *);
void treeprint (FILE *, struct llist *, int);
void listprint (FILE *, struct llist *, int);
int listcount (struct llist *);
int input (int, int);
void removedash (char []);

enum comptypes { Dunno, Amiga, Apple, Atari, IBM_DOS, IBM_OS2, IBM_Linux, IBM_Windows, Other };

struct llist {
	char softwarename [40];
	int count;
	struct llist *next;
	struct llist *nextright;
};

struct llist *start [maxcomp];

char ibmname[maxibm] [13] = {      /* names of IBM computers */
	"8086",
	"286",
	"386",
	"486",
	"pentium",
	"unspecified"
			    };

char ibmname2[maxibm] [6] = {        /* Alternate names  */
	"xt",
	"at",
	"80386",
	"80486",
	"-",
	"-"
			    };

char baudcodes[maxbaud] [10] = {            /* Possible baud rates */
	"v.fc",
	"Turbopep",
	"v.32terbo",
	"Pep",
	"v.32bis",                          /* In order from highest to lowest */
	"HST",                              /* bps speed */
	"v.32",
	"v.22bis",
	"v.22",
	"v.23",
	"v.21"
			  };

long int approxbps[maxbaud] = {      /* bps speed achievable using the above */
	28800,                       /* May or may not be correct */
	23000,
	19200,
	14400,
	14400,
	9600,
	9600,
	2400,
	1200,
	1200,
	300
				   };


char compname [maxcomp] [15] =
{
	"",           /* 0 isn't used */
	"Amiga",
	"Apple",
	"Atari",
	"IBM - MS-DOS",
	"IBM - OS/2",
	"IBM - Linux",
	"IBM - Windows",
	"Other"
};

int ibm[maxibm];                /* Number of IBM computers of each type read */
int baud[maxbaud], bps [maxbaud];
time_t tstart, tend;            /* For difftime() timing of program */

int computer;                   /* Computer type read (number) */
int cab, bars, sortorder;
int noofbars;
int clone, unspec;
int fulltime, parttime;
int test;
char comptype [81];       /* Computer type read (name) */

void main (int argc, char *argv[])
{
	#define maxfiles 9
	FILE *fp;
	char temp[81];
	int loop, loop2, result, quit = no, readfiles = no;
	long size [maxfiles];                   /* File sizes in K */

	/* BBS listings file names */
	char files [maxfiles] [12] = {
	"BBSFALL.TXT",
	"BBSFACT.TXT",
	"BBSFNSW.TXT",
	"BBSFNT.TXT",
	"BBSFQLD.TXT",
	"BBSFSA.TXT",
	"BBSFTAS.TXT",
	"BBSFVIC.TXT",
	"BBSFWA.TXT"
		      };
	/* use command line argument of 'test' to turn test mode on */
	test = (argc > 1 && strcmp(*(argv + 1),"test") == 0);
	initialise ();
	intro();
	testing();
	time( &tstart );     /* Used to time the program  */
	for (loop = 0; loop < maxfiles; loop ++)
	{
		testing();
		noofbars = 0;
		cab = 0;
		size [loop] = getsizes(files[loop]);
		fp = fopen(files[loop],"r");
		if (fp == NULL)
		{
			sprintf (temp, "%s%s\r\n","Unable to find ",files[loop]);
			cputs(temp);
			fclose(fp);
			continue;
		}
		readfiles = yes;
		clrscr();
		sprintf(temp,"\r\n%s%s%s%lu%s\r\n","Reading ",files[loop],".  File size = ",size[loop]," K.  Press S to stop.");
		cputs (temp);
		if (bars < 3)
			drawit();
		quit = no;
		while (quit == no)
		{
			quit = checkfors();
			if (quit == yes)
			{
				fclose(fp);
				continue;
			}
			strcpy(temp,"");
			result = fscanf(fp,"%[^\n]",temp);
			if (result == EOF)
			{
				if (bars < 3)
					last();
				fclose(fp);
				quit = yes;
				continue;
			}
			getc (fp);
			cab ++;
			if (bars < 3)
				show(size[loop]);
			if (strnicmp(temp,"   Computer: ",13) == 0)
			{
				computer = Dunno;
				strcpy(comptype,temp);
				foundcomp(&temp[13],1);
				continue;
			}
			if (strnicmp(temp,"         OS: ",13) == 0)
			{
				foundcomp(&temp[13],2);
				if (test == yes && computer == Other)
					printf("Unknown computer type:  %s\r\n",comptype);
				continue;
			}
			if (strnicmp(temp," BBSoftware: ",13) == 0)
			{
				foundsw (&temp[13]);
				continue;
			}
			if (strnicmp(temp,"       Baud: ",13) == 0)
			{
				foundbaud(temp);
				continue;
			}
			if (strnicmp(temp,"      Hours: ",13) == 0)
				foundhours(temp);

		}

	}
	if (readfiles == yes)
		writeresults();
	else
		printf("Program aborted\r\n");
	_setcursortype (_NORMALCURSOR);
}

void foundcomp(char original[],int attemptno)
{
	#define noofcomp 19
	int loop, foundit;
	int pdest;
	/* Finding out what comp are used */
	/* by checking the 'computer' and 'os' lines */
	/* If attemptno = 1 then it's checking 'computer' */
	/* If attemptno = 2 then it's checking 'os' */

	char keywords [noofcomp] [8] =
	{
			"amiga",
			"mac",
			"apple",
			"atari",
			"ms-dos",
			"ms dos",
			"pc-dos",
			"pc-dos",
			"dr dos",
			"286",
			"xt",
			"386",
			"at",
			"pentium",
			"486",
			"ibm",
			"os/2",
			"linux",
			"windows"
	};

	/* These are defined in a global enum statement */
	int comptypeno [noofcomp] = { Amiga,Apple,Apple,Atari,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_DOS,IBM_OS2,IBM_Linux, IBM_Windows};


	for (loop = 0; loop < noofcomp; loop ++)
	{
		pdest = find(original,keywords [loop]);
		if (pdest == yes)
		{
			computer = comptypeno [loop];
			break;
		}
	}
	if (computer == Dunno && attemptno == 2)    /* If unable to find computer type */
		computer = Other;

	if (computer == IBM_DOS || computer == IBM_OS2 || computer == IBM_Linux || computer == IBM_Windows)
	{
		if (attemptno == 1)
		{
			pdest = find(original,"clone");   /* Check for clones as opposed to real IBM's */
			if (pdest)
				clone ++;
		}
		if (attemptno == 2)
		{
			foundit = no;
			for (loop = 0; loop < maxibm; loop ++)
			{
				pdest = find(comptype,ibmname[loop]);
				if (pdest == 0)
					pdest = find(comptype,ibmname2[loop]);
				if (pdest)
				{
					ibm[loop] ++;
					foundit = yes;
					break;
				}
			}
		}
		if (foundit == no)
			ibm[maxibm - 1] ++;           /* 'unspecified' */
	}

}


void foundsw(char software[])
{
	#define maxstuff 13
	int loop;
	int len, len2;

	char stuff [maxstuff] [10] =
	{
		"!",
		"+",
		"/2",
		" //",
		"/ST",
		" 16 Line",
		" 32 Line",
		" 64 Line",
		"/Pro",
		" Pro",
		" BBS",
		" FX",
		" XLT"
	};

	len = strlen (software);
	for (loop = 0; loop < maxstuff; loop ++)     /* Remove stuff from end of software name */
	{
		len2 = len - strlen (stuff [loop]);
		if (strlen (software) > len2 && stricmp (stuff [loop], &software[len2]) == 0)
		{
			len -= strlen (stuff [loop]);
			software [len2] = '\0';
			break;
		}
	}
	removedash (software);

	if (software[0] >= 'a' && software[0] <= 'z')      /* Captialise first letter  */
		software[0] -= 'a' - 'A';
    /*    if (test == yes)
		printf("Software sent to list function is *%s*\r\n",software);
     */
	/* Find or add software in the tree or linked list */
	if (test == yes && strlen (software) < 1)
		printf("Error!  Strlen (software) = 0\r\n");

	if (sortorder == alphabetical)
		start [computer] = addtotree (start [computer], software);
	else
		start [computer] = addtolist (start [computer], software);
}

void foundbaud(char string[])
{
	int loopa;
	int foundit;

	foundit = no;
	for (loopa = 0; loopa < maxbaud; loopa++)
	{
		if ( find (string,baudcodes[loopa]) == yes)
		{
			if (foundit == no)
			{
				bps[loopa] ++;
				foundit = yes;
			}
			baud [loopa] ++;
		}
	}
	if (test == yes && foundit == no)
		printf("Unknown baud code(s):  %s\r\n",string);

}

void foundhours(char string[])
{
	if (stricmp(string,"      Hours: 24 Hours") == 0)
		fulltime ++;
	else
		parttime ++;
}

int checkfors()     /* Check to see if S has been pressed */
{
	int key;
	fflush(stdin);
	if ( (key = kbhit()) != NULL)
	{
		key = getch();
		key = (key == 's' || key == 'S') ? yes : no;
	}
	return(key);
}

void intro()
{
	clrscr();
	cprintf("\r\n");
	cputs("       BBS list reading program by C.A.B.  January 1994.  Public domain.\r\n");
	cputs("\r\n1  Status bar showing progress.\r\n2  Percentage read shown.\r\n3  Neither.\r\n");
	cputs ("\r\nChoose 1, 2 or 3:  ");
	bars = input (1,3);
	clrscr();
	cputs("\r\n\r\n1  Show software names in alphabetical order.\r\n2  Show software in order of popularity.\r\n");
	cputs ("\r\nChoose 1 or 2:  ");
	sortorder = input (1,2);
	clrscr();

}

void writeresults()
{
	FILE *fp;
	int loopa, loopb, loopc, loopd, total, timetaken, howmanybbs = 0;

	int comp, noofibm = 0, realibm = 0;
	char temp[81];

	testing();
	clrscr();
	time( &tend );
	timetaken = (int) difftime(tend,tstart);
	sprintf(temp,"Finished.  Time taken:  %i seconds.\r\n",timetaken);
	cputs(temp);
	cputs("\r\nWriting results to the file BBSLIST.DAT.  Please wait.\r\n");

	fp = fopen("bbslist.dat","w");
	if (fp == NULL)
	{
		cputs("Error.   Unable to write BBSLIST.DAT\r\n");
		cputs ("Program aborted\r\n");
		exit (0);
	}
	else
	{
		for (loopa = 0; loopa < maxbaud; loopa ++)      /* Find out how many BBS's have been read */
			howmanybbs += bps[loopa];
		for (comp = 1; comp < maxcomp; comp ++)
		{
			fprintf(fp," \nComputer:  %s\n \n",compname[comp]);
			total = listcount (start [comp]);  /* Add up total no. of BBS's run on this comp type */

			if (sortorder == alphabetical)
				treeprint (fp, start [comp], total);
			else
				listprint (fp, start [comp], total);

			fprintf(fp,"Number of BBS's run on this type of computer:  %i\n",total);

		}
		noofibm = 0;
		for (loopc = 0; loopc < maxibm; loopc ++)
			noofibm += ibm[loopc];

		fprintf(fp," \nTotal no. of BBS's read:  %i\n",howmanybbs);
		fprintf(fp," \nNumber of BBS's run on IBM's or clones: %i\n",noofibm);
		realibm = noofibm - clone;

		for (loopc = 0; loopc < maxibm; loopc ++)
		{
			strcat(ibmname[loopc],":");
			sprintf(temp,"%-14s",ibmname[loopc]);
			writepercent(fp,ibm[loopc],noofibm,temp);
		}

		writepercent(fp,realibm,noofibm," \nNumber run on IBM's:   ");
		writepercent(fp,clone,noofibm,"Number run on clones:  ");

		fprintf(fp," \nData modulation methods:\n");
		for (loopc = (maxbaud - 1); loopc >= 0; loopc--)
		{
			strcat(baudcodes[loopc],":");
			sprintf(temp,"%-22s",baudcodes[loopc]);
			writepercent(fp,baud[loopc],howmanybbs,temp);
		}

		fprintf (fp, " \nHighest connection speed (bps):\n");
		for (loopc = (maxbaud - 1); loopc >= 0; loopc--)
		{
			/* Two baud codes might have the same bps rate */
			if (loopc == 0 || approxbps [loopc] != approxbps [loopc - 1])
			{
				sprintf(temp,"%-22li",approxbps[loopc]);
				writepercent(fp,bps[loopc],howmanybbs,temp);
			}
			if (approxbps [loopc] == approxbps [loopc - 1])
				bps [loopc - 1] += bps [loopc];
		}

		fprintf(fp," \nNumber of fulltime and parttime BBS's:\n");
		writepercent(fp,fulltime,howmanybbs,"Fulltime:  ");
		writepercent(fp,parttime,howmanybbs,"Parttime:  ");
		fprintf (fp,"\n");
	}
	fclose(fp);

	showfile();
	textbackground(BLACK);
	textcolor(LIGHTGRAY);
	clrscr();
	_setcursortype( _NORMALCURSOR );
	exit(0);
}


int find(char astring[],char whattofind[])
{
	char *pdest;
	char subset [81];
	char string [81];
	int num;
	int result;

	strcpy (subset,strlwr(whattofind));
	strcpy (string,strlwr(astring));
	num = strlen (subset);
	result = no;

	if (strnicmp(string,subset,num) == 0)
		return(yes);
	pdest = strstr(string,subset);
	if (pdest != NULL)
		result = yes;
	return(result);

}

long getsizes(char file [])
{
	int handle;
	long size;

	handle = open( file, O_TEXT | O_RDONLY );
	if (handle != -1)
	{
		size = filelength( handle );
		size = (size > 1023) ? size /= 1024 : 1;

	}
	close(handle);
	return (size);
}

void drawit()
{
		if (bars == yes)
		{
			cputs("Ŀ\r\n");
			cputs("                    \r\n");
			cputs("\r\n");
			cputs("0         50        100%");
			gotoxy(2,4);
		}
		else
		{
			cputs("   %");
		}
}

void show(float size)
{
		static float last = 0;
		float prog;
		char temp[50];
		int loopa;
		float lines = (size * 1000 / 32);

		if (cab > lines)
			cab = lines;
		prog = (int) ( (cab/lines) * 100 );
		if (prog == last)
			return;
		last = prog;
		loopa = (int) (prog/5);
		if ( prog / 5 == (int) prog / 5)
		{
			if (bars == yes && noofbars < 20)
			{
				cprintf("");
			}
			if (bars != yes)
			{
				gotoxy(2,3);
				loopa *= 5;
				itoa(loopa,temp,10);
				cprintf("%2s",temp);
			}
			noofbars ++;
		}

}

void last()
{
	if (bars == yes)
	{
		while(noofbars < 20)
		{
			cprintf("");
			noofbars ++;
		}
	}
	else
	{
		gotoxy(2,3);
		cputs("100");
	}
}



void showfile()
{
	FILE *fp;
	char temp [81];
	char temp2 [2];
	int loopa = 0;
	int result;

	fp = fopen("bbslist.dat","rt");
	if (fp != NULL)
	{
		while (1)
		{
			strcpy(temp,"");
			if ( fgets(temp,80,fp) == NULL)
				break;
			loopa ++;
			if (loopa > 22 || strncmp(temp,"Computer:  ",11) == 0 || strncmp(temp,"Total no. of BBS's read:",24) == 0 || strncmp (temp,"Data modulation",15) == 0)
			{
				cputs("\r\nPress enter.  ");
				fflush(stdin);
				gets(temp2);
				clrscr();
				loopa = 0;
			}
			cprintf("%s\r",temp);
		}
	}
	fclose(fp);
	cputs("\r\nPress enter.  ");
	fflush(stdin);
	gets(temp);
	clrscr();


}

void writepercent(FILE *fp, int small,int large,char string[])
{

	unsigned long percent;

	percent = (unsigned long) small * 100 / large;
	fprintf(fp,string);
	fprintf(fp,"%4i (",small);
	fprintf(fp,"%3lu percent)\n",percent );

}

void initialise ()
{
	int loop, loop2;
	clone = computer = fulltime = parttime = 0;

	for (loop = 0; loop < maxcomp; loop ++)
		start [loop] = NULL;
	for (loop = 0; loop < maxbaud; loop ++)
		baud[loop] = 0;
	for (loop = 0; loop < maxibm; loop ++)
		ibm[loop] = 0;

}

void testing()
{
	int a;
	/*
	if (test == yes)
		for (a = 1; a < maxcomp; a++)
			printf("comp %i = *%s*\n",a,compname[a]);

	 */
}

struct llist *addtotree (struct llist *branch, char *software)
{
	int cond;

	if (test == yes)
		printf ("addtotree, software = %s\n",software);

	if (branch == NULL)       /* Adding a new branch to the binary tree */
	{
		branch = (struct llist *) malloc(sizeof (struct llist) );
		if (branch == NULL)   /* Out of memory trying to create new node*/
		{
			puts("Not enough memory");
			exit (1);
		}
		strcpy (branch->softwarename,software);
		branch->count = 1;
		branch->next = branch->nextright = NULL;

	}
	else if ((cond = stricmp (branch->softwarename, software)) == 0)   /* Found the software name */
		branch->count ++;
	else if (cond < 0)
		branch->next = addtotree (branch->next, software); /* left branch */
	else
		branch->nextright = addtotree (branch->nextright, software); /* right branch */
	return (branch);

}

struct llist *addtolist (struct llist *node, char software[])
{
    /*    if (test == yes)
		printf("Software name received is *%s* computer type is %i\n",software,computer);
     */
	if (node == NULL)       /* Adding a new node to the linked list */
	{
		node = (struct llist *) malloc(sizeof (struct llist) );
		if (node == NULL)   /* Out of memory trying to create new node*/
		{
			puts("Not enough memory");
			exit (1);
		}
		strcpy (node->softwarename,software);
		node->count = 1;
		node->next = NULL;
	  /*      if (test == yes)
			printf("just added node: software sent = *%s* software stored = *%s* count = *%i*\n",software,node->softwarename,node->count);
	  */
	}
	else if (stricmp (node->softwarename, software) == 0)   /* Found the software name */
	{
		node->count ++;
	   /*     if (test == yes)
			printf("Found %s count now = *%i*\n",software, node->count);
	   */
	}
	else
		node->next = addtolist (node->next, software); /* left node */
	return (node);
}

void treeprint (FILE *fp, struct llist *branch, int total)
{          /* will print the software names in alphabetical order */

	if (test == yes)
		printf("In treeprint\n");
	if (branch != NULL)
	{
		treeprint (fp, branch->nextright, total);
		fprintf(fp,"Software:  %-21s",branch->softwarename);
		writepercent(fp, branch->count,total,"no. of users:  ");
		treeprint (fp, branch->next, total);

	}
}

void listprint (FILE *fp, struct llist *node, int total)
{        /* Will print the software names in order of number of users (highest to lowest) */

	struct llist *pointer, *temp;   /* temp is a temporary pointer to a node */
	int highest;

	if (test == yes)
		printf("In listprint\n");

	if (node)    /* If list is not empty */
	do
	{
		pointer = node;    /* Start at the start of the list */
		highest = 0;       /* highest no of users found so far */
		do
		{
			if (pointer->count > highest)     /* Keep track of one with highest no. of users */
			{
				highest = pointer->count;
				temp = pointer;
			}
			pointer = pointer->next;          /* go to the next node in the list */

		}  while (pointer != NULL);    /* While not at end of list (marked by NULL) */

		if (highest)   /* if the node with highest no of users has more than 0  */
		{
			fprintf(fp,"Software:  %-21s",temp->softwarename);
			writepercent(fp, temp->count,total,"no. of users:  ");
			temp->count = 0;                 /* set it to 0 to mark it as been done */

		}

	}
	while (highest);    /* do until all nodes have been done */

}


int listcount (struct llist *start)
{
	int total = 0;
	if (test == yes)
		printf("In listcount\n");
	if (start != NULL)
	{
		total = start->count;
       /*         if (test == yes)
			printf("Looking at %s total = %i count = %i\n",start->softwarename,total,start->count);
       */
		total += listcount (start->next);
		if (test == yes)
			printf("total after recursive call = %i\n",total);
		if (sortorder == alphabetical)
			total += listcount (start->nextright);
	}
	return (total);
}

int input (int min, int max)
{
	int loopa = 0, result;

	do
	{
		fflush(stdin);
		while (loopa == 0)
			loopa = kbhit();
		result = getch();
	}
	while (result < min + '0' || result > max + '0');
	result -= '0';
	return (result);
}

void removedash (char string [])
{              /* Remove any -'s or spaces in the string, and also anything */
	       /* after a / or an & or a comma                              */

	       int i,j;
	       for (i = j = 0; string[i] != '\0' && string[i] != '/' && string[i] != ',' && string[i] != '&'; i ++)
		if (string [i] != '-' && string [i] != ' ')
			string [j ++] = string [i];
		string [j] = '\0';
}
